@@ -13,13 +13,18 @@ from facebook_api.models import FacebookGraphIDModel, FacebookGraphManager, MAST
 from facebook_applications.models import Application
 from facebook_users.models import User
 from facebook_pages.models import Page
+from m2m_history.fields import ManyToManyHistoryField
 from datetime import datetime
+import dateutil.parser
 import logging
 import time
 import re
 
 log = logging.getLogger('facebook_posts')
 
+class UnknownResourceType(Exception):
+    pass
+
 def get_or_create_from_small_resource(resource):
     '''
     Return instance of right type based on dictionary resource from Facebook API Graph
@@ -27,8 +32,10 @@ def get_or_create_from_small_resource(resource):
     keys = sorted(resource.keys())
     defaults = dict(resource)
     del defaults['id']
-    if keys == ['category','id','name']:
+    if keys == ['category','id','name'] or keys == ['category','category_list','id','name']:
         # resource is a page
+        if 'category_list' in defaults:
+            del defaults['category_list']
         return Page.objects.get_or_create(graph_id=resource['id'], defaults=defaults)[0]
     elif keys == ['id','name']:
         # resource is a user
@@ -36,7 +43,8 @@ def get_or_create_from_small_resource(resource):
     elif keys == ['id','name','namespace']:
         # resource is a application
         return Application.objects.get_or_create(graph_id=resource['id'], defaults=defaults)[0]
-        pass
+    else:
+        raise UnknownResourceType("Resource with strange keys: %s" % keys)
 
 class PostFacebookGraphManager(FacebookGraphManager):
 
@@ -47,7 +55,7 @@ class PostFacebookGraphManager(FacebookGraphManager):
 
     @atomic
     @fetch_all(return_all=update_count_and_get_posts, always_all=True, paging_next_arg_name='until')
-    def fetch_page_wall(self, page, limit=1000, offset=0, until=None, since=None, **kwargs):
+    def fetch_page_wall(self, page, limit=250, offset=0, until=None, since=None, **kwargs):
         '''
         Arguments:
          * until|since - timestamp or datetime
@@ -108,8 +116,11 @@ class FacebookLikableModel(models.Model):
         if response:
             log.debug('response objects count=%s, limit=%s, after=%s' % (len(response.data), limit, kwargs.get('after')))
             for resource in response.data:
-                user = get_or_create_from_small_resource(resource)
-                ids += [user.pk]
+                try:
+                    user = get_or_create_from_small_resource(resource)
+                    ids += [user.pk]
+                except UnknownResourceType:
+                    continue
 
         return User.objects.filter(pk__in=ids), response
 
@@ -119,8 +130,6 @@ class Post(FacebookGraphIDModel, FacebookLikableModel):
         verbose_name = 'Facebook post'
         verbose_name_plural = 'Facebook posts'
 
-    like_users = models.ManyToManyField(User, related_name='like_posts')
-
     # in API field called `from`
     author_json = fields.JSONField(null=True, help_text='Information about the user who posted the message') # object containing the name and Facebook id of the user who posted the message
     owners_json = fields.JSONField(null=True, help_text='Profiles mentioned or targeted in this post') # Contains in data an array of objects, each with the name and Facebook id of the user
@@ -140,7 +149,7 @@ class Post(FacebookGraphIDModel, FacebookLikableModel):
 
     picture = models.TextField(help_text='If available, a link to the picture included with this post')
     source = models.TextField(help_text='A URL to a Flash movie or video file to be embedded within the post')
-    link = models.URLField(max_length=500, help_text='The link attached to this post')
+    link = models.URLField(max_length=1000, help_text='The link attached to this post')
     icon = models.URLField(max_length=500, help_text='A link to an icon representing the type of this post')
 
     name = models.TextField(help_text='The name of the link')
@@ -172,6 +181,9 @@ class Post(FacebookGraphIDModel, FacebookLikableModel):
 
     shares_count = models.IntegerField(default=0)
 
+    like_users = ManyToManyHistoryField(User, related_name='like_posts')
+    shares_users = ManyToManyHistoryField(User, related_name='shares_posts')
+
     objects = models.Manager()
     remote = PostFacebookGraphManager()
 
@@ -198,7 +210,10 @@ class Post(FacebookGraphIDModel, FacebookLikableModel):
 
         if self.owners.count() == 0 and self.owners_json:
             for owner_json in self.owners_json:
-                owner = get_or_create_from_small_resource(owner_json)
+                try:
+                    owner = get_or_create_from_small_resource(owner_json)
+                except UnknownResourceType:
+                    continue
                 if owner:
                     self._external_links_to_add += [('owners', PostOwner(post=self, owner=owner))]
 
@@ -223,6 +238,55 @@ class Post(FacebookGraphIDModel, FacebookLikableModel):
 
         return Comment.objects.filter(pk__in=ids), response
 
+    def update_count_and_get_shares_users(self, instances, *args, **kwargs):
+        self.shares_users = instances
+        self.shares_count = instances.count()
+        self.save()
+        return instances
+
+    @atomic
+    @fetch_all(return_all=update_count_and_get_shares_users, paging_next_arg_name='after')
+    def fetch_shares(self, limit=1000, **kwargs):
+        '''
+        Retrieve and save all shares of post
+        '''
+        ids = []
+        response = graph('%s/sharedposts' % self.graph_id.split('_')[1], **kwargs)
+
+        if response:
+
+            timestamps = dict([(int(post['from']['id']), dateutil.parser.parse(post['created_time'])) for post in response.data])
+            ids_new = timestamps.keys()
+            # becouse we should use local pk, instead of remote, remove it after pk -> graph_id
+            ids_current = map(int, User.objects.filter(pk__in=self.shares_users.get_query_set(only_pk=True).using(MASTER_DATABASE).exclude(time_from=None)).values_list('graph_id', flat=True))
+            ids_add = set(ids_new).difference(set(ids_current))
+            ids_add_pairs = []
+            ids_remove = set(ids_current).difference(set(ids_new))
+
+            log.debug('response objects count=%s, limit=%s, after=%s' % (len(response.data), limit, kwargs.get('after')))
+            for post in response.data:
+                graph_id = int(post['from']['id'])
+                if sorted(post['from'].keys()) == ['id','name']:
+                    try:
+                        user = get_or_create_from_small_resource(post['from'])
+                        ids += [user.pk]
+                        # this id in add list and still not in add_pair (sometimes there are duplicates)
+                        if graph_id in ids_add and graph_id not in map(lambda i:i[0], ids_add_pairs):
+                            ids_add_pairs += [(graph_id, user.pk)]  # becouse we should use local pk, instead of remote
+                    except UnknownResourceType:
+                        continue
+
+            m2m_model = self.shares_users.through
+
+            # remove old shares without time_from
+            self.shares_users.get_query_set_through().filter(time_from=None).delete()
+
+            # add new shares
+            get_share_date = lambda id: timestamps[id] if id in timestamps else self.created_time
+            m2m_model.objects.bulk_create([m2m_model(**{'user_id': pk, 'post_id': self.pk, 'time_from': get_share_date(graph_id)}) for graph_id, pk in ids_add_pairs])
+
+        return User.objects.filter(pk__in=ids), response
+
     def save(self, *args, **kwargs):
         # set exactly Page or User contentTypes, not a child
         for field_name in ['author']:
@@ -280,8 +344,6 @@ class Comment(FacebookGraphIDModel, FacebookLikableModel):
         verbose_name = 'Facebook comment'
         verbose_name_plural = 'Facebook comments'
 
-    like_users = models.ManyToManyField(User, related_name='like_comments')
-
     post = models.ForeignKey(Post, related_name='comments')
     author_json = fields.JSONField(null=True, help_text='Information about the user who posted the comment') # object containing the name and Facebook id of the user who posted the message
 
@@ -295,6 +357,8 @@ class Comment(FacebookGraphIDModel, FacebookLikableModel):
     can_remove = models.BooleanField(default=False)
     user_likes = models.BooleanField(default=False)
 
+    like_users = ManyToManyHistoryField(User, related_name='like_comments')
+
     objects = models.Manager()
     remote = FacebookGraphManager()
 
